"""Test custom relations for JSON parser"""
import pytest

DATA = {
    "data": [
        {
            "name": "Jane",
            "age": 28,
            "lab_no": 123,
        },
        {
            "name": "John",
            "age": 32,
            "lab_no": 345,
        },
    ]
}


DATA_NULL = {
    "data": [
        {
            "name": "",
            "age": 28,
            "lab_no": 0,
        },
        {
            "name": "John",
            "age": 32,
            "lab_no": 345,
        },
    ]
}

DATA_SUBGRAPHS = {
    "data": [
        {
            "name": "specimen_1",
            "youngs_modulus": 100,
            "material": "material_1",
            "kitem": "kitem_1",
            "tensile_strength": [100, 200],
        },
        {
            "name": "specimen_2",
            "youngs_modulus": "200-300",
            "material": "material_2",
            "kitem": ["kitem_2", "kitem_3"],
            "tensile_strength": [200, 300],
        },
        {
            "name": "specimen_3",
            "youngs_modulus": [300, 400],
            "material": "material_3",
            "kitem": "kitem_4",
            "tensile_strength": [300, 400],
        },
    ]
}

MAPPING_SUBGRAPHS = [
    {
        "iri": "https://w3id.org/steel/ProcessOntology/Specimen",
        "suffix": "name",
        "source": "data[*]",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "youngs_modulus",
                "relation": "https://w3id.org/steel/ProcessOntology/hasYoungsModulus",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/YoungsModulus",
                    "unit": "GPa",
                },
            },
            {
                "object_location": "material",
                "relation": "https://w3id.org/steel/ProcessOntology/hasMaterial",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/Material",
                    "value_relation": "https://w3id.org/steel/ProcessOntology/hasIdentifier",
                },
            },
            {
                "object_location": "name",
                "relation": "https://w3id.org/steel/ProcessOntology/hasIdentifier",
            },
            {
                "object_location": "kitem",
                "relation": "https://w3id.org/steel/ProcessOntology/hasKitem",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/Kitem/",
                    "concatenate": True,
                },
            },
            {
                "object_location": "tensile_strength[0]",
                "relation": "https://w3id.org/steel/ProcessOntology/hasMinimumTensileStrength",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/TensileStrength",
                    "unit": "MPa",
                    "suffix": "tensile_strength_min",
                },
            },
            {
                "object_location": "tensile_strength[1]",
                "relation": "https://w3id.org/steel/ProcessOntology/hasMaximumTensileStrength",
                "object_type": {
                    "iri": "https://w3id.org/steel/ProcessOntology/TensileStrength",
                    "unit": "MPa",
                    "suffix": "tensile_strength_max",
                },
            },
        ],
    }
]

EXPECTED_SUBGRAPHS = """
@prefix ns1: <https://w3id.org/steel/ProcessOntology/> .
@prefix ns2: <http://qudt.org/schema/qudt/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#specimen_1> a ns1:Specimen ;
    ns1:hasIdentifier "specimen_1"^^xsd:string ;
    ns1:hasMaterial <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_1> ;
    ns1:hasYoungsModulus <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_1> ;
    ns1:hasMaximumTensileStrength <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_max_specimen_1> ;
    ns1:hasMinimumTensileStrength <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_min_specimen_1> ;
    ns1:hasKitem <https://w3id.org/steel/ProcessOntology/Kitem/kitem_1> .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#specimen_2> a ns1:Specimen ;
    ns1:hasIdentifier "specimen_2"^^xsd:string ;
    ns1:hasMaterial <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_2> ;
    ns1:hasYoungsModulus <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_2> ;
    ns1:hasMaximumTensileStrength <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_max_specimen_2> ;
    ns1:hasMinimumTensileStrength <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_min_specimen_2> ;
    ns1:hasKitem <https://w3id.org/steel/ProcessOntology/Kitem/kitem_2>, <https://w3id.org/steel/ProcessOntology/Kitem/kitem_3> .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#specimen_3> a ns1:Specimen ;
    ns1:hasIdentifier "specimen_3"^^xsd:string ;
    ns1:hasMaterial <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_3> ;
    ns1:hasYoungsModulus <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_3> ;
    ns1:hasMaximumTensileStrength <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_max_specimen_3> ;
    ns1:hasMinimumTensileStrength <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_min_specimen_3> ;
    ns1:hasKitem <https://w3id.org/steel/ProcessOntology/Kitem/kitem_4> .


<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_1> a ns1:Material ;
    ns1:hasIdentifier "material_1"^^xsd:string .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_2> a ns1:Material ;
    ns1:hasIdentifier "material_2"^^xsd:string .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#Material_specimen_3> a ns1:Material ;
    ns1:hasIdentifier "material_3"^^xsd:string .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_1> a ns1:YoungsModulus ;
    ns2:hasUnit "http://qudt.org/vocab/unit/GigaPA"^^xsd:anyURI ;
    ns2:value 100 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_2> a ns1:YoungsModulus ;
    ns2:hasUnit "http://qudt.org/vocab/unit/GigaPA"^^xsd:anyURI ;
    ns2:value "200-300"^^xsd:string .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#YoungsModulus_specimen_3> a ns1:YoungsModulus ;
    ns2:hasUnit "http://qudt.org/vocab/unit/GigaPA"^^xsd:anyURI ;
    ns2:value 300, 400 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_max_specimen_1> a ns1:TensileStrength ;
    ns2:hasUnit "http://qudt.org/vocab/unit/MegaPA"^^xsd:anyURI ;
    ns2:value 200 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_max_specimen_2> a ns1:TensileStrength ;
    ns2:hasUnit "http://qudt.org/vocab/unit/MegaPA"^^xsd:anyURI ;
    ns2:value 300 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_max_specimen_3> a ns1:TensileStrength ;
    ns2:hasUnit "http://qudt.org/vocab/unit/MegaPA"^^xsd:anyURI ;
    ns2:value 400 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_min_specimen_1> a ns1:TensileStrength ;
    ns2:hasUnit "http://qudt.org/vocab/unit/MegaPA"^^xsd:anyURI ;
    ns2:value 100 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_min_specimen_2> a ns1:TensileStrength ;
    ns2:hasUnit "http://qudt.org/vocab/unit/MegaPA"^^xsd:anyURI ;
    ns2:value 200 .

<https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#tensile_strength_min_specimen_3> a ns1:TensileStrength ;
    ns2:hasUnit "http://qudt.org/vocab/unit/MegaPA"^^xsd:anyURI ;
    ns2:value 300 .
"""

EXPECTED = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
@prefix ns3: <https://w3id.org/steel/ProcessOntology/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ns2:Jane a ns1:Operator ;
            foaf:name "Jane"^^xsd:string ;
            foaf:age "28"^^xsd:integer ;
            ns3:hasLaboratory "123"^^xsd:integer .

ns2:John a ns1:Operator ;
            foaf:name "John"^^xsd:string ;
            foaf:age "32"^^xsd:integer ;
            ns3:hasLaboratory "345"^^xsd:integer .
"""

EXPECTED_DATATYPE = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
@prefix ns3: <https://w3id.org/steel/ProcessOntology/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ns2:Jane a ns1:Operator ;
            foaf:name "Jane"^^xsd:string ;
            foaf:age "28"^^xsd:string ;
            ns3:hasLaboratory "123"^^xsd:string .

ns2:John a ns1:Operator ;
            foaf:name "John"^^xsd:string ;
            foaf:age "32"^^xsd:string ;
            ns3:hasLaboratory "345"^^xsd:string .
"""

MAPPING_WILDCARD = [
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "name",
        "source": "data[*]",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "name",
                "relation": "http://xmlns.com/foaf/0.1/name",
            },
            {
                "object_location": "age",
                "relation": "http://xmlns.com/foaf/0.1/age",
            },
            {
                "object_location": "lab_no",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
            },
        ],
    }
]

MAPPING_INDEX = [
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "data[0].name",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "data[0].name",
                "relation": "http://xmlns.com/foaf/0.1/name",
            },
            {
                "object_location": "data[0].age",
                "relation": "http://xmlns.com/foaf/0.1/age",
            },
            {
                "object_location": "data[0].lab_no",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
            },
        ],
    },
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "data[1].name",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "data[1].name",
                "relation": "http://xmlns.com/foaf/0.1/name",
            },
            {
                "object_location": "data[1].age",
                "relation": "http://xmlns.com/foaf/0.1/age",
            },
            {
                "object_location": "data[1].lab_no",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
            },
        ],
    },
]


MAPPING_WILDCARD_DATATYPE = [
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "name",
        "source": "data[*]",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "name",
                "relation": "http://xmlns.com/foaf/0.1/name",
                "object_data_type": "string",
            },
            {
                "object_location": "age",
                "relation": "http://xmlns.com/foaf/0.1/age",
                "object_data_type": "string",
            },
            {
                "object_location": "lab_no",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
                "object_data_type": "string",
            },
        ],
    }
]

MAPPING_INDEX_DATATYPE = [
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "data[0].name",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "data[0].name",
                "relation": "http://xmlns.com/foaf/0.1/name",
                "object_data_type": "string",
            },
            {
                "object_location": "data[0].age",
                "relation": "http://xmlns.com/foaf/0.1/age",
                "object_data_type": "string",
            },
            {
                "object_location": "data[0].lab_no",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
                "object_data_type": "string",
            },
        ],
    },
    {
        "iri": "https://w3id.org/emmo/domain/characterisation-methodology/chameo#Operator",
        "suffix": "data[1].name",
        "suffix_from_location": True,
        "custom_relations": [
            {
                "object_location": "data[1].name",
                "relation": "http://xmlns.com/foaf/0.1/name",
                "object_data_type": "string",
            },
            {
                "object_location": "data[1].age",
                "relation": "http://xmlns.com/foaf/0.1/age",
                "object_data_type": "string",
            },
            {
                "object_location": "data[1].lab_no",
                "relation": "https://w3id.org/steel/ProcessOntology/hasLaboratory",
                "object_data_type": "string",
            },
        ],
    },
]

BASE_IRI = (
    "https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation"
)

ADDITIONAL_TRIPLES = """
@prefix : <http://abox-namespace-placeholder.org/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .

:John foaf:knows :Jane .
"""

EXPECTED_ADDITIONAL_TRIPLES = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
@prefix ns3: <https://w3id.org/steel/ProcessOntology/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ns2:Jane a ns1:Operator ;
            foaf:name "Jane"^^xsd:string ;
            foaf:age "28"^^xsd:integer ;
            ns3:hasLaboratory "123"^^xsd:integer .


ns2:John a ns1:Operator ;
            foaf:name "John"^^xsd:string ;
            foaf:age "32"^^xsd:integer ;
            ns3:hasLaboratory "345"^^xsd:integer ;
            foaf:knows ns2:Jane .
"""


EXPECTED_NULL = """
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix ns1: <https://w3id.org/emmo/domain/characterisation-methodology/chameo#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix ns2: <https://w3id.org/emmo/domain/domain-nanoindentation/nanoindentation#> .
@prefix ns3: <https://w3id.org/steel/ProcessOntology/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

ns2:Operator a ns1:Operator ;
           foaf:age "28"^^xsd:integer ;
           ns3:hasLaboratory "0"^^xsd:integer .


ns2:John a ns1:Operator ;
            foaf:name "John"^^xsd:string ;
            foaf:age "32"^^xsd:integer ;
            ns3:hasLaboratory "345"^^xsd:integer .
"""


@pytest.mark.parametrize("mapping", [MAPPING_WILDCARD, MAPPING_INDEX])
def test_pipeline_json_custom_relations(mapping) -> None:
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA,
        mapping=mapping,
        parser=Parser.json,
        config={
            "base_iri": BASE_IRI,
            "separator": "#",
            "prefix_name": "nanoindentation",
            "suppress_file_description": True,
        },
    )
    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED)

    assert pipeline.graph.isomorphic(expected_graph)


@pytest.mark.parametrize(
    "mapping", [MAPPING_WILDCARD_DATATYPE, MAPPING_INDEX_DATATYPE]
)
def test_pipeline_json_custom_relations_datatype(mapping) -> None:
    """Test with custom relations and datatypes"""

    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA,
        mapping=mapping,
        parser=Parser.json,
        config={
            "base_iri": BASE_IRI,
            "separator": "#",
            "prefix_name": "nanoindentation",
            "suppress_file_description": True,
        },
    )

    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_DATATYPE)

    assert pipeline.graph.isomorphic(expected_graph)


def test_pipeline_json_custom_relations_additional_triples() -> None:
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser

    pipeline = Data2RDF(
        raw_data=DATA,
        mapping=MAPPING_WILDCARD,
        parser=Parser.json,
        config={
            "base_iri": BASE_IRI,
            "separator": "#",
            "prefix_name": "nanoindentation",
            "suppress_file_description": True,
        },
        additional_triples=ADDITIONAL_TRIPLES,
    )
    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_ADDITIONAL_TRIPLES)

    assert pipeline.graph.isomorphic(expected_graph)


def test_pipeline_json_custom_relations_additional_triples_null() -> None:
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser
    from data2rdf.warnings import MappingMissmatchWarning

    with pytest.warns(UserWarning, match="Concept with for ") as warnings:
        pipeline = Data2RDF(
            raw_data=DATA_NULL,
            mapping=MAPPING_WILDCARD,
            parser=Parser.json,
            config={
                "base_iri": BASE_IRI,
                "separator": "#",
                "prefix_name": "nanoindentation",
                "suppress_file_description": True,
            },
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_NULL)

    assert pipeline.graph.isomorphic(expected_graph)


def test_pipeline_json_custom_relations_quantity_graph() -> None:
    from rdflib import Graph

    from data2rdf import Data2RDF, Parser
    from data2rdf.warnings import ParserWarning

    with pytest.warns(ParserWarning, match="Cannot") as warnings:
        pipeline = Data2RDF(
            raw_data=DATA_SUBGRAPHS,
            mapping=MAPPING_SUBGRAPHS,
            parser=Parser.json,
            config={
                "base_iri": BASE_IRI,
                "separator": "#",
                "suppress_file_description": True,
            },
        )

    warnings = [
        warning for warning in warnings if warning.category == ParserWarning
    ]
    assert len(warnings) == 1

    expected_graph = Graph()
    expected_graph.parse(data=EXPECTED_SUBGRAPHS)

    assert pipeline.graph.isomorphic(expected_graph)



"""Basic data2rdf models"""

import json
from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import (
    AnyUrl,
    BaseModel,
    ConfigDict,
    Field,
    field_validator,
    model_validator,
)
from rdflib import Graph

from data2rdf.config import Config


class RelationType(str, Enum):
    """Relation Type of TBox modellings"""

    ANNOTATION_PROPERTY = "annotation_property"
    DATA_PROPERTY = "data_property"
    OBJECT_PROPERTY = "object_property"
    PROPERTY = "property"


class BaseConfigModel(BaseModel):
    """Basic model for holding the data2rdf config"""

    config: Config = Field(
        default_factory=Config, description="Configuration object"
    )
    model_config = ConfigDict(exclude={"config"})

    def __str__(self) -> str:
        """Pretty print the model"""
        values = ",\n".join(
            [
                f"\t{key}={value}"
                for key, value in self.__dict__.items()
                if key not in self.model_config.get("exclude")
            ]
        )
        return f"{self.__class__.__name__}(\n{values})"

    def __repr__(self) -> str:
        """Pretty print the model"""
        return str(self)

    @field_validator("config")
    @classmethod
    def validate_config(cls, value: Union[Dict[str, Any], Config]) -> Config:
        """Validate configuration"""
        if isinstance(value, dict):
            value = Config(**value)
        return value


class BasicConceptMapping(BaseConfigModel):
    """Basic mapping for a concept in a file"""

    key: Optional[str] = Field(
        None, description="Key/column of the concept in the file"
    )


class BasicGraphModel(BasicConceptMapping):
    """Basic model for merging data with mappings to become a graph"""

    @property
    @abstractmethod
    def json_ld(self) -> Dict[str, Any]:
        """Return dict for json-ld of graph"""

    @property
    def graph(self) -> Graph:
        """Return graph object based on json-ld"""
        graph = Graph(identifier=self.config.graph_identifier)
        graph.parse(data=json.dumps(self.json_ld), format="json-ld")
        return graph


class BasicSuffixModel(BaseConfigModel):
    """Pydantic BaseModel for suffix and type of a class instance"""

    iri: Union[str, AnyUrl, List[Union[str, AnyUrl]]] = Field(
        ..., description="Ontological class related to this concept"
    )
    suffix: Optional[str] = Field(
        None,
        description="""Optional suffix of the individual
        which will be constructed. If not set, the suffix of the iri
        of the ontological class will be taken""",
        validate_default=True,
    )

    @field_validator("iri")
    @classmethod
    def validate_iri(cls, value: Union[AnyUrl, List[AnyUrl]]) -> AnyUrl:
        """Make sure that there are not blank spaces in the IRI"""
        if not isinstance(value, list):
            value = AnyUrl(str(value).strip())
        else:
            value = [AnyUrl(str(iterable).strip()) for iterable in value]
        return value

    @model_validator(mode="after")
    @classmethod
    def validate_suffix(
        cls,
        self: "BasicSuffixModel",
    ) -> "BasicSuffixModel":
        """Return suffix for individal"""

        if isinstance(self.iri, list) and self.suffix is None:
            raise TypeError("If the iri is a list, the suffix must be set ")

        self.suffix = (
            self.suffix or str(self.iri).split(self.config.separator)[-1]
        )
        return self



"""Data2RDF base model for parsers"""

import json
import warnings
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

from rdflib import Graph

from data2rdf.config import Config
from data2rdf.modes import PipelineMode

from .utils import load_mapping_file

from pydantic import (  # isort:skip
    BaseModel,
    Field,
    PrivateAttr,
    field_validator,
    model_validator,
    AnyUrl,
)


if TYPE_CHECKING:
    import pandas as pd

    from data2rdf import BasicConceptMapping


class BaseParser(BaseModel):
    """Basic Parser for any data file and mode"""

    raw_data: Union[str, bytes, Dict[str, Any], List[Dict[str, Any]]] = Field(
        ...,
        description="""
        In case of a csv: `str` with the file path or the content of the file itself.
        In case of a json file: `dict` for the content of the file of `str` for the file content or file path.
        In case of an excel file: `btyes` for the content or `str` for the file path""",
    )
    mapping: Union[str, List[Any]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )

    dropna: bool = Field(
        False,
        description="Drop all rows where ONLY NaN and None occur in the dataframe.",
    )

    config: Config = Field(
        default_factory=Config, description="Configuration object"
    )

    @field_validator("config")
    @classmethod
    def validate_config(cls, value: Union[Dict[str, Any], Config]) -> Config:
        """Validate configuration"""
        if isinstance(value, dict):
            value = Config(**value)
        return value


class AnyBoxBaseParser(BaseParser):
    """Basic parser for A Box or T Box producing an RDF"""

    @property
    @abstractmethod
    def json_ld(self) -> Dict[str, Any]:
        """Return dict for json-ld for the graph"""

    @property
    @abstractmethod
    def mapping_model(self) -> "BaseParser":
        """Pydantic model for validating mapping.
        Must be a subclass of `ABoxBaseParser` or `TBoxBaseParser`.
        """

    @classmethod
    @abstractmethod
    def _run_parser(
        cls, self, datafile: Any, mapping: "Dict[str, BaseParser]"
    ) -> None:
        """Class method for running parser. The `datafile` argument is the
        object returned by the `_load_data_file` method and the `mapping` is
        a dictionary of the keys/columns from the data file mapped to instances
        of the `mapping_model` property of this class."""

    @classmethod
    @abstractmethod
    def _load_data_file(cls, self: "BaseParser") -> "Dict[str, Any]":
        """Class method for loading data file"""

    @property
    def graph(self) -> "Graph":
        """Return RDF Graph from the parsed data."""
        graph = Graph(identifier=self.config.graph_identifier)
        graph.parse(data=json.dumps(self.json_ld), format="json-ld")
        return graph

    @model_validator(mode="after")
    @classmethod
    def run_parser(cls, self: "BaseParser") -> "BaseParser":
        """
        Runs the parser for the given data file and mapping.

        This function is a class method that takes in a `self` parameter, which is an instance of the `BaseParser` class.
        It loads the data file using the `_load_data_file` method and loads the mapping file using the `load_mapping_file` function.
        It then runs the parser using the `_run_parser` method and returns the parsed `BaseParser` instance.

        Args:
            self (BaseParser): The instance of the `BaseParser` class.

        Returns:
            BaseParser: The parsed `BaseParser` instance.
        """

        datafile: Any = cls._load_data_file(self)
        mapping: "Dict[str, BaseParser]" = load_mapping_file(
            self.mapping, self.config, self.mapping_model
        )
        cls._run_parser(self, datafile, mapping)
        return self


class TBoxBaseParser(AnyBoxBaseParser):
    """Basic Parser for TBox mode"""

    suffix_location: str = Field(
        ...,
        description="""Key/column name/reference to the locaton in the data file
        where the suffix of the ontological class to be created.""",
    )

    rdfs_type_location: Optional[str] = Field(
        None,
        description="""Key/column name/reference to location in the data file
        where the rdfs:type for the concepts is defined.""",
    )

    version_info: Optional[str] = Field(
        None, description="Version of the ontology"
    )

    ontology_iri: Optional[Union[str, AnyUrl]] = Field(
        None, description="General IRI of the ontology."
    )

    ontology_title: Optional[str] = Field(
        None, description="Title of the ontology"
    )

    authors: Optional[List[str]] = Field(
        None, description="Name of the authors contributing to the ontology."
    )

    fillna: Optional[Any] = Field(
        "", description="Value to fill NaN values in the parsed dataframe."
    )

    _classes: Any = PrivateAttr()

    @property
    def classes(self) -> "List[BasicConceptMapping]":
        """Return list object with class models"""
        return self._classes


class ABoxBaseParser(AnyBoxBaseParser):
    """Basic Parser for ABox mode"""

    _general_metadata: Any = PrivateAttr()
    _dataframe_metadata: Any = PrivateAttr()
    _dataframe: Any = PrivateAttr()

    @property
    def general_metadata(self) -> "List[BasicConceptMapping]":
        """Return list object with general metadata"""
        return self._general_metadata

    @property
    def dataframe_metadata(self) -> "List[BasicConceptMapping]":
        """Return list object with general metadata"""
        return self._dataframe_metadata

    @property
    def dataframe(self) -> "pd.DataFrame":
        """Return times series found in the data as pd.DataFrame"""
        return self._dataframe

    @property
    def plain_metadata(self) -> List[Dict[str, Any]]:
        message = """
        `plain_metadata` is deprecated and will be removed in a future version.
        Use the `to_dict()` instead."""
        warnings.warn(message, DeprecationWarning)
        return self.to_dict()

    def to_dict(
        self, schema: Callable = None
    ) -> "Union[Dict[str, Any], List[Dict[str, Any]]]":
        """
        Return general metadata as a list of dictionaries.

        The list contains dictionaries, where the key is the label of the metadata,
        and the value is a dictionary with the keys 'label' and 'value'. If the
        metadata has a measurement unit associated with it, the dictionary will
        also contain the key 'measurement_unit' with the value of the measurement
        unit.

        If the schema parameter is provided, it will be used to transform the
        metadata list. The schema should be a callable which takes the list of
        metadata dictionaries and returns the transformed metadata.

        If no schema is provided, the function will return a dictionary where the
        keys are the labels of the metadata, and the values are the dictionaries
        from the list.

        :param schema: A callable which takes a list of dictionaries and returns
            the transformed metadata.
        :return: A dictionary or list of dictionaries with the metadata.
        """
        metadata = []
        for metadatum in self.general_metadata:
            prop = {
                "label": metadatum.suffix,
                "value": metadatum.value,
                "relation_mapping": {
                    "class_iri": str(metadatum.iri),
                },
            }
            if hasattr(metadatum, "measurement_unit"):
                prop[
                    "measurement_unit"
                ] = metadatum.measurement_unit.model_dump(exclude={"config"})
            metadata.append(prop)
        if not isinstance(schema, type(None)):
            metadata = schema(metadata)
        else:
            metadata = {datum.get("label"): datum for datum in metadata}
        return metadata


class BaseFileParser(BaseParser):
    """Base model for data files which can be run in abox
    or tbox mode. The respective `ABoxBaseParser` and
    `TBoxBaseParser` must be set as properties for this model.
    The childclasses of this `BaseFileParser` will be directly used by the main
    `Data2RDF` class later."""

    mode: PipelineMode = Field(
        PipelineMode.ABOX, description="Run parser in ABox or TBox mode."
    )

    parser_args: Dict[str, Any] = Field(
        {},
        description="A dict with specific arguments for the parser. Is passed to the parser as kwargs.",
    )

    _abox: Any = PrivateAttr()
    _tbox: Any = PrivateAttr()

    @property
    @abstractmethod
    def media_type(self) -> "Union[str, AnyUrl]":
        """IANA Media type definition of the resources to be parsed."""

    @property
    @abstractmethod
    def _abox_parser(self) -> "ABoxBaseParser":
        """Childclass of `ABoxBaseParser` for the specific `BaseFileParser`."""

    @property
    @abstractmethod
    def _tbox_parser(self) -> "TBoxBaseParser":
        """Childclass of `TBoxBaseParser` for the specific `BaseFileParser`."""

    @property
    def abox(self) -> "ABoxBaseParser":
        """Return instance of the `abox_parser` after model validation"""
        return self._abox

    @property
    def tbox(self) -> "TBoxBaseParser":
        """Return instance of the `tbox_parser` after model validation"""
        return self._tbox

    @model_validator(mode="after")
    @classmethod
    def execute_parser(cls, self: "BaseFileParser") -> "BaseFileParser":
        """
        Validates the parser model and executes the parser based on the specified mode.

        Args:
            self: An instance of the BaseFileParser class.

        Returns:
            An instance of the BaseFileParser class with the parser executed.
        """
        arguments = {
            "mapping": self.mapping,
            "raw_data": self.raw_data,
            "config": self.config,
            **self.parser_args,
        }
        if self.mode == PipelineMode.ABOX:
            self._abox = self._abox_parser(**arguments)
        elif self.mode == PipelineMode.TBOX:
            self._tbox = self._tbox_parser(**arguments)
        else:
            raise TypeError(f"Operating mode not understood: {self.mode}")
        return self

    @property
    def plain_metadata(self) -> Dict[str, Any]:
        """Metadata as flat json - without units and iris.
        Useful e.g. for the custom properties of the DSMS."""
        if self.mode == PipelineMode.ABOX:
            return self.abox.plain_metadata
        else:
            raise NotImplementedError(
                "`plain_metadata` is not available in `tbox`-mode."
            )

    def to_dict(self, schema: Callable = None) -> "List[Dict[str, Any]]":
        """Return list of general metadata as DSMS custom properties"""
        if self.mode == PipelineMode.ABOX:
            return self.abox.to_dict(schema=schema)
        else:
            raise NotImplementedError(
                "`to_dict()` is not available in `tbox`-mode."
            )

    @property
    def general_metadata(self) -> "List[BasicConceptMapping]":
        """Return list object with general metadata"""
        if self.mode == PipelineMode.ABOX:
            return self.abox.general_metadata
        else:
            raise NotImplementedError(
                "`general_metadata` is not available in `tbox`-mode."
            )

    @property
    def dataframe_metadata(self) -> "List[BasicConceptMapping]":
        """Return dataframe metadata"""
        if self.mode == PipelineMode.ABOX:
            return self.abox.dataframe_metadata
        else:
            raise NotImplementedError(
                "`dataframe_metadata` is not available in `tbox`-mode."
            )

    @property
    def dataframe(self) -> "Dict[str, Any]":
        """Return dataframe"""
        if self.mode == PipelineMode.ABOX:
            return self.abox.dataframe
        else:
            raise NotImplementedError(
                "`dataframe` is not available in `tbox`-mode."
            )

    @property
    def graph(self) -> Graph:
        """Return RDFlib Graph"""
        if self.mode == PipelineMode.ABOX:
            return self.abox.graph
        else:
            return self.tbox.graph

    @property
    def json_ld(self) -> "Dict[str, Any]":
        """Return JSON LD representation of graph"""
        if self.mode == PipelineMode.ABOX:
            return self.abox.json_ld
        else:
            return self.tbox.json_ld



"""data2rdf unit test for models"""

import pytest

normal_config = {"graph_identifier": "https://www.example.org"}
bad_config = {"graph_identifier": "https://www.example.org", "foorbar": 123}

unit_string = "mm"
unit_iri = "http://qudt.org/vocab/unit/MilliM"


@pytest.mark.parametrize("config", [normal_config, bad_config])
def test_quantity_model(config) -> None:
    from rdflib import Graph

    from data2rdf import QuantityGraph

    expected = """@prefix fileid: <https://www.example.org/> .
    @prefix qudt: <http://qudt.org/schema/qudt/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

    fileid:test a <https://example.org/test> ;
        qudt:hasUnit "http://qudt.org/vocab/unit/MilliM"^^xsd:anyURI ;
        qudt:value "0.1"^^xsd:float .
    """
    expected_graph = Graph()
    expected_graph.parse(format="turtle", data=expected)

    model = QuantityGraph(
        value=0.1,
        key="test",
        unit="mm",
        iri="https://example.org/test",
        config=config,
    )

    assert model.graph.isomorphic(expected_graph)
    assert str(model.graph.identifier) == config["graph_identifier"]

    assert model.measurement_unit.iri == "http://qudt.org/vocab/unit/MilliM"
    assert model.measurement_unit.symbol == "mm"
    assert model.measurement_unit.label == "Millimetre"
    assert model.measurement_unit.namespace == "http://qudt.org/vocab/unit"


@pytest.mark.parametrize("unit", [unit_string, unit_iri])
def test_valued_quantity(unit):
    from rdflib import Graph

    from data2rdf import QuantityGraph

    expected = """@prefix fileid: <https://www.example.org/> .
    @prefix qudt: <http://qudt.org/schema/qudt/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

    fileid:test a <https://example.org/test> ;
        qudt:hasUnit "http://qudt.org/vocab/unit/MilliM"^^xsd:anyURI ;
        qudt:value "0.1"^^xsd:float .
    """
    expected_graph = Graph()
    expected_graph.parse(format="turtle", data=expected)

    model = QuantityGraph(
        value=0.1, key="test", unit=unit, iri="https://example.org/test"
    )

    assert model.graph.isomorphic(expected_graph)

    assert model.measurement_unit.iri == "http://qudt.org/vocab/unit/MilliM"
    assert model.measurement_unit.symbol == "mm"
    assert model.measurement_unit.label == "Millimetre"
    assert model.measurement_unit.namespace == "http://qudt.org/vocab/unit"


def test_bad_with_blank_space():
    from rdflib import Graph

    from data2rdf import QuantityGraph

    expected = """@prefix fileid: <https://www.example.org/> .
    @prefix qudt: <http://qudt.org/schema/qudt/> .
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

    fileid:test a <https://example.org/test> ;
        qudt:hasUnit "http://qudt.org/vocab/unit/MilliM"^^xsd:anyURI ;
        qudt:value "0.1"^^xsd:float .
    """
    expected_graph = Graph()
    expected_graph.parse(format="turtle", data=expected)

    model = QuantityGraph(
        value=0.1, key="test", unit="mm", iri="https://example.org/test     "
    )

    assert model.graph.isomorphic(expected_graph)

    assert model.measurement_unit.iri == "http://qudt.org/vocab/unit/MilliM"
    assert model.measurement_unit.symbol == "mm"
    assert model.measurement_unit.label == "Millimetre"
    assert model.measurement_unit.namespace == "http://qudt.org/vocab/unit"



"""Data2RDF parser utilities"""

import json
import re
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from data2rdf import Config
from data2rdf.models.graph import ClassTypeGraph
from data2rdf.warnings import MappingMissmatchWarning

from data2rdf.models.mapping import (  # isort:skip
    RelationType,
)

ALLOW_PATTERN = r"[^a-zA-Z0-9\-._$[*\[\]]*"

if TYPE_CHECKING:
    from typing import Any, Dict, List, Union

    from data2rdf.models.mapping import TBoxBaseMapping
    from data2rdf.parsers.base import TBoxBaseParser


def load_mapping_file(
    mapping: "Union[str, List[Dict[str, Any]]]",
    config: "Config" = Config(),
    model_callable: "Any" = dict,
) -> "List[Any]":
    """
    Load a mapping file and transform its contents into a dictionary format.

    Parameters:
        mapping (Union[str, List[Dict[str, Any]]]): Path to the mapping file (either a string representing the file path or a list containing the mapping directly).
        config (Config, optional): Configuration settings for loading the file. Defaults to Config().
        model_callable (Any, optional): Callable object used to transform each row of the mapping file into the desired format. Defaults to dict.

    Returns:
        List[Any]: A list containing the loaded mapping data.

    Raises:
        TypeError: If the `mapping` parameter is not of type `str` or `dict`, or if the file type for mapping is not supported.

    Note:
        - For Excel files (.xlsx), the 'sameas' sheet is read from the Excel file.
        The contents are then transformed into a dictionary where each row corresponds to a key-value pair.
        - For CSV files (.csv) please take care of the correct separator to be set in the config under `mapping_csv_separator`.
        - For JSON files (.json), the entire file is loaded as a dictionary.
        - If `mapping` is already a dictionary, it is returned as is.

    """
    if not isinstance(mapping, (str, list)):
        raise TypeError(
            f"""Mapping file must be of type `{str}` or `{list}`,
            not `{type(mapping)}`."""
        )
    if isinstance(mapping, str):
        if mapping.endswith("xlsx"):
            mapping_df = pd.read_excel(
                mapping,
                sheet_name="sameas",
                engine="openpyxl",
            )
            mapping_df.fillna("", inplace=True)
            mapping_df = mapping_df.apply(lambda s: s.str.replace('"', ""))
            model = [row.to_dict() for n, row in mapping_df.iterrows()]

        elif mapping.endswith("json"):
            with open(mapping, encoding=config.encoding) as file:
                model = json.load(file)
        elif mapping.endswith("csv"):
            mapping_df = pd.read_csv(
                mapping,
                sep=config.mapping_csv_separator,
                encoding=config.encoding,
            )
            mapping_df.fillna("", inplace=True)
            mapping_df = mapping_df.apply(lambda s: s.str.replace('"', ""))
            model = [row.to_dict() for n, row in mapping_df.iterrows()]
        else:
            raise TypeError("File type for mapping not supported!")

        result = [model_callable(**row) for row in model]
    if isinstance(mapping, list):
        result = mapping
    return result


def _strip_unit(symbol: str, char_list: "List[str]") -> str:
    for char in char_list:
        symbol = symbol.strip(char)
    return symbol


def _make_tbox_classes(
    self: "TBoxBaseParser",
    df: "pd.DataFrame",
    mapping: "List[TBoxBaseMapping]",
) -> None:
    self._classes = []
    mapping = {model.key: model for model in mapping}
    if hasattr(self, "header_length"):
        skipped = df[self.header_length - 1 :]
    else:
        skipped = df
    for n, row in skipped.iterrows():
        annotations = []
        datatypes = []
        objects = []
        properties = []

        for key, model in mapping.items():
            try:
                value = row[key]
                if isinstance(value, float) and np.isnan(value):
                    value = self.fillna
                if value:
                    relation_mapping = {
                        "value": value,
                        "relation": model.relation,
                        "datatype": model.datatype,
                    }
                    if model.relation_type == RelationType.ANNOTATION_PROPERTY:
                        annotations.append(relation_mapping)
                    if model.relation_type == RelationType.DATA_PROPERTY:
                        datatypes.append(relation_mapping)
                    if model.relation_type == RelationType.OBJECT_PROPERTY:
                        objects.append(relation_mapping)
                    if model.relation_type == RelationType.PROPERTY:
                        properties.append(relation_mapping)
                else:
                    warnings.warn(
                        f"Data for key `{key}` does not exist in row {n}.",
                        MappingMissmatchWarning,
                    )
            except KeyError:
                warnings.warn(
                    f"Data for key `{key}` does not exist in row {n}.",
                    MappingMissmatchWarning,
                )

        if self.rdfs_type_location:
            rdfs_type = row[self.rdfs_type_location]
            if isinstance(rdfs_type, type(None)) or (
                isinstance(rdfs_type, float) and np.isnan(rdfs_type)
            ):
                warnings.warn(
                    f"Data for key `{self.rdfs_type_location}` does not exist in row {n}.",
                    MappingMissmatchWarning,
                )
                rdfs_type = "owl:Class"
        else:
            rdfs_type = "owl:Class"
        subgraph = ClassTypeGraph(
            rdfs_type=rdfs_type,
            suffix=row[self.suffix_location],
            annotation_properties=annotations,
            object_properties=objects,
            data_properties=datatypes,
            rdfs_properties=properties,
            config=self.config,
        )
        self._classes.append(subgraph)


def _make_tbox_json_ld(model: "TBoxBaseParser") -> "Dict[str, Any]":
    ontology_iri = model.ontology_iri or model.config.base_iri
    classes = [model.json_ld for model in model.classes]
    if not model.config.exclude_ontology_title:
        if model.ontology_title:
            title = {"dcterms:title": model.ontology_title}
        else:
            title = {}
        if model.authors:
            authors = {
                "dcterms:creator": [
                    {"@type": "foaf:Person", "foaf:name": author}
                    for author in model.authors
                ]
            }
        else:
            authors = {}
        if model.version_info:
            version = {"owl:versionInfo": model.version_info}
        else:
            version = {}

        classes += [
            {
                "@id": str(ontology_iri),
                "@type": "owl:Ontology",
                **title,
                **authors,
                **version,
            },
        ]
    return {
        "@context": {
            "owl": "http://www.w3.org/2002/07/owl#",
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "dcterms": "http://purl.org/dc/terms/",
            "foaf": "http://xmlns.com/foaf/spec/",
        },
        "@graph": classes,
    }


def _value_exists(value: "Any") -> bool:
    """
    Check if a given value exists and is valid.

    This method verifies if the input value is neither None, an empty string,
    nor a NaN (Not a Number) value.

    Args:
        value (Any): The value to be checked.

    Returns:
        bool: True if the value exists and is valid, otherwise False.
    """
    return pd.notnull(value) and value != ""


def _check_jsonpath(expression):
    """
    Adjusts the JSONPath expression for compatibility.

    This function checks the provided JSONPath expression against a predefined
    pattern and modifies it for compatibility by enclosing each segment of the
    path in double quotes if any matches are found. This ensures that the
    expression is properly formatted for further processing.

    Args:
        expression (str): The JSONPath expression to be checked and adjusted.

    Returns:
        str: The adjusted JSONPath expression with segments enclosed in double quotes
             if any matches were found, otherwise returns the original expression.
    """

    matches = re.findall(ALLOW_PATTERN, expression)
    matches = [match for match in matches if match]
    if matches:
        print(matches)
        splitted = expression.split(".")
        expression = ".".join([f'"{exp}"' for exp in splitted])
    return expression



"""DSMS Warnings"""


class MappingMissmatchWarning(UserWarning):
    """A missmatch warning if a value or a concept was not corrently mapped"""


class ParserWarning(UserWarning):
    """A warning raised for a specific context set for a parser"""


class QUDTMappingWarning(UserWarning):
    """A warning raised for a specific context set for a QUDT mapping"""



"""General data2rdf utils"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from data2rdf.config import Config


def make_prefix(config: "Config") -> str:
    if not str(config.base_iri).endswith(config.separator):
        prefix = str(config.base_iri) + config.separator
    else:
        prefix = str(config.base_iri)
    return prefix


def split_namespace(iri: str) -> tuple[str, str]:
    """
    Split the given iri into a namespace and a localname.

    Args:
        iri: The iri to split.

    Returns:
        A tuple of the namespace and the localname.
    """
    if "#" in iri:
        return iri.split("#")[0]
    else:
        return "/".join(iri.split("/")[:-1])



"""Data2RDF models"""

from .graph import PropertyGraph, QuantityGraph
from .mapping import ABoxBaseMapping, BasicConceptMapping

__all__ = [
    "QuantityGraph",
    "PropertyGraph",
    "ABoxBaseMapping",
    "BasicConceptMapping",
]



"""CSV Parser pytest"""

import json
import os

import pytest

from ..utils import as_non_dsms_schema, dsms_schema, remove_ids, sort_entries

test_folder = os.path.dirname(os.path.abspath(__file__))
working_folder = os.path.join(test_folder, "input")
output_folder = os.path.join(test_folder, "output")

mapping_folder = os.path.join(working_folder, "mapping")
raw_data = os.path.join(working_folder, "data", "DX56_D_FZ2_WR00_43.TXT")
expected = os.path.join(output_folder, "output_csv_parser.ttl")

parser_args = {
    "metadata_sep": "\t",
    "dataframe_sep": "\t",
    "metadata_length": 20,
}

metadata = {
    "sections": [
        {
            "entries": [
                {
                    "label": "TestingFacility",
                    "value": "institute_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/TestingFacility"
                    },
                },
                {
                    "label": "ProjectNumber",
                    "value": "123456",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/ProjectNumber"
                    },
                },
                {
                    "label": "ProjectName",
                    "value": "proj_name_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/ProjectName"
                    },
                },
                {
                    "label": "TimeStamp",
                    "value": "44335.4",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/TimeStamp"
                    },
                },
                {
                    "label": "MachineData",
                    "value": "maschine_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/MachineData"
                    },
                },
                {
                    "label": "ForceMeasuringDevice",
                    "value": "Kraftaufnehmer_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/ForceMeasuringDevice"
                    },
                },
                {
                    "label": "DisplacementTransducer",
                    "value": "Wegaufnehmer_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/DisplacementTransducer"
                    },
                },
                {
                    "label": "TestStandard",
                    "value": "ISO-XX",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/TestStandard"
                    },
                },
                {
                    "label": "Material",
                    "value": "Werkstoff_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Material"
                    },
                },
                {
                    "label": "SpecimenType",
                    "value": "Probentyp_1",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SpecimenType"
                    },
                },
                {
                    "label": "Tester",
                    "value": "abc",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Tester"
                    },
                },
                {
                    "label": "SampleIdentifier-2",
                    "value": "Probentyp_2",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SampleIdentifier-2"
                    },
                },
                {
                    "label": "OriginalGaugeLength",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 80,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/OriginalGaugeLength"
                    },
                },
                {
                    "label": "ParallelLength",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 120,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/ParallelLength"
                    },
                },
                {
                    "label": "SpecimenThickness",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 1.55,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SpecimenThickness"
                    },
                },
                {
                    "label": "SpecimenWidth",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM",
                        "label": "Millimetre",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm",
                    },
                    "value": 20.04,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/SpecimenWidth"
                    },
                },
                {
                    "label": "TestingRate",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MilliM-PER-SEC",
                        "label": "Millimetre per Second",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "mm/s",
                    },
                    "value": 0.1,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/TestingRate"
                    },
                },
                {
                    "label": "Preload",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/MegaPA",
                        "label": "Megapascal",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "MPa",
                    },
                    "value": 2,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Preload"
                    },
                },
                {
                    "label": "Temperature",
                    "measurement_unit": {
                        "iri": "http://qudt.org/vocab/unit/DEG_C",
                        "label": "degree Celsius",
                        "namespace": "http://qudt.org/vocab/unit",
                        "symbol": "°C",
                    },
                    "value": 22,
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Temperature"
                    },
                },
                {
                    "label": "Remark",
                    "value": "",
                    "relation_mapping": {
                        "class_iri": "https://w3id.org/steel/ProcessOntology/Remark"
                    },
                },
            ],
            "name": "General",
        },
    ],
}

columns = [
    "TestTime",
    "StandardForce",
    "Extension",
    "Elongation",
    "AbsoluteCrossheadTravel",
    "WidthChange",
]

normal_config = {"graph_identifier": "https://www.example.org"}
bad_config = {"graph_identifier": "https://www.example.org", "foorbar": 123}


def test_csv_parser_bad_mapping() -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import CSVParser

    parser = CSVParser(
        raw_data=raw_data,
        mapping=os.path.join(mapping_folder, "bad_tensile_test_mapping.json"),
        parser_args=parser_args,
    )
    expected_graph = Graph()
    expected_graph.parse(expected)

    assert len(parser.general_metadata) == 20
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 5734

    assert parser.graph.isomorphic(expected_graph)


def test_csv_parser_no_match_in_mapping() -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import CSVParser
    from data2rdf.warnings import MappingMissmatchWarning

    with pytest.warns(
        MappingMissmatchWarning, match="No match found"
    ) as warnings:
        parser = CSVParser(
            raw_data=os.path.join(
                working_folder, "data", "BAD_DX56_D_FZ2_WR00_43.TXT"
            ),
            mapping=os.path.join(mapping_folder, "tensile_test_mapping.json"),
            parser_args={
                "metadata_sep": "\t",
                "dataframe_sep": "\t",
                "metadata_length": 21,
            },
        )

    missmatches = [
        warning
        for warning in warnings
        if warning.category == MappingMissmatchWarning
    ]
    assert len(missmatches) == 1

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert len(parser.general_metadata) == 20
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 5734

    assert parser.graph.isomorphic(expected_graph)


@pytest.mark.parametrize("config", [normal_config, bad_config])
def test_csv_parser_config(config) -> None:
    from rdflib import Graph

    from data2rdf.parsers import CSVParser

    parser = CSVParser(
        raw_data=raw_data,
        mapping=os.path.join(mapping_folder, "tensile_test_mapping.json"),
        parser_args=parser_args,
        config=config,
    )
    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)
    assert str(parser.graph.identifier) == config["graph_identifier"]
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)


@pytest.mark.parametrize("extension", ["xlsx", "json", "csv", dict])
def test_parser_csv(extension) -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import CSVParser

    if isinstance(extension, str):
        mapping = os.path.join(
            mapping_folder, f"tensile_test_mapping.{extension}"
        )

    else:
        path = os.path.join(mapping_folder, "tensile_test_mapping.json")
        with open(path, encoding="utf-8") as file:
            mapping = json.load(file)

    parser = CSVParser(
        raw_data=raw_data,
        mapping=mapping,
        parser_args=parser_args,
    )

    assert len(parser.general_metadata) == 20
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 5734

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )
    assert sort_entries(parser.to_dict()) == as_non_dsms_schema(metadata)


@pytest.mark.parametrize("input_kind", ["path", "content"])
def test_parser_csv_input(input_kind) -> None:
    from rdflib import Graph

    from data2rdf.models import PropertyGraph, QuantityGraph
    from data2rdf.parsers import CSVParser

    if input_kind == "path":
        input_obj = raw_data
    elif input_kind == "content":
        with open(raw_data, encoding="utf-8") as file:
            input_obj = file.read()

    parser = CSVParser(
        raw_data=input_obj,
        mapping=os.path.join(mapping_folder, "tensile_test_mapping.json"),
        parser_args=parser_args,
    )

    assert len(parser.general_metadata) == 20
    for row in parser.general_metadata:
        assert isinstance(row, QuantityGraph) or isinstance(row, PropertyGraph)

    assert len(parser.dataframe_metadata) == 6
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 6
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)
    for name, column in parser.dataframe.items():
        assert len(column) == 5734

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert parser.graph.isomorphic(expected_graph)

    assert remove_ids(parser.to_dict(schema=dsms_schema)) == sort_entries(
        metadata
    )

    assert sort_entries(parser.to_dict()) == as_non_dsms_schema(metadata)



"""CSV Parser for data2rdf"""

import os
import warnings
from io import StringIO
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin

import numpy as np
import pandas as pd
from pydantic import AliasChoices, Field

from data2rdf.models.graph import PropertyGraph, QuantityGraph
from data2rdf.utils import make_prefix
from data2rdf.warnings import MappingMissmatchWarning, ParserWarning

from .base import ABoxBaseParser, BaseFileParser, TBoxBaseParser
from .utils import _make_tbox_classes, _make_tbox_json_ld, _strip_unit

from data2rdf.models.mapping import (  # isort:skip
    ABoxBaseMapping,
    TBoxBaseMapping,
)


def _replace(value: Optional[str], to_be_replaced: List[str]) -> Any:
    """Replace char in string"""
    if isinstance(value, str):
        for char in to_be_replaced:
            value = value.replace(char, "")
    return value


def _load_data_file(self: "Union[CSVTBoxParser, CSVABoxParser]") -> StringIO:
    """Load csv file"""
    if isinstance(self.raw_data, str):
        if os.path.isfile(self.raw_data):
            with open(self.raw_data, encoding=self.config.encoding) as file:
                content = StringIO(file.read())
        else:
            content = StringIO(self.raw_data)
    else:
        raise TypeError(
            f"`raw_data` must be of type `str`, not `{type(self.raw_data)}`"
        )
    return content


class CSVTBoxParser(TBoxBaseParser):
    """
    CSV file parser in tbox mode
    """

    # OVERRIDE
    mapping: Union[str, List[TBoxBaseMapping]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )
    column_sep: Optional[str] = Field(",", description="Data column separator")
    header_length: int = Field(
        1, description="Length of the header of the CSV sheet", ge=1
    )

    fillna: Optional[Any] = Field(
        "", description="Value to fill NaN values in the parsed dataframe."
    )

    # OVERRIDE
    @property
    def mapping_model(self) -> TBoxBaseMapping:
        """TBox Mapping Model for CSV Parser"""
        return TBoxBaseMapping

    # OVERRIDE
    @property
    def json_ld(self) -> "Dict[str, Any]":
        """Make the json-ld if pipeline is in abox-mode"""
        return _make_tbox_json_ld(self)

    # OVERRIDE
    @classmethod
    def _run_parser(
        cls,
        self: "CSVTBoxParser",
        datafile: StringIO,
        mapping: "List[TBoxBaseMapping]",
    ) -> None:
        """
        Class method for running the CSVTBoxParser. This method reads a CSV file
        into a pandas DataFrame and then uses the provided mapping to create TBox
        classes.

        Parameters:
            self (CSVTBoxParser): The instance of the parser.
            datafile (StringIO): The CSV file to be parsed.
            mapping (List[TBoxBaseMapping]): The list of mappings to be applied.

        Returns:
            None
        """

        df = pd.read_csv(datafile, sep=self.column_sep)
        _make_tbox_classes(self, df, mapping)

    # OVERRIDE
    @classmethod
    def _load_data_file(cls, self: "CSVTBoxParser") -> StringIO:
        """Load CSV file"""
        return _load_data_file(self)


class CSVABoxParser(ABoxBaseParser):
    """
    CSV file parser in abox mode
    """

    metadata_sep: Optional[str] = Field(
        None, description="Metadata column separator"
    )
    metadata_length: int = Field(..., description="Length of the metadata")
    dataframe_sep: Optional[str] = Field(
        None,
        description="Column separator of the dataframe header",
        alias=AliasChoices("dataframe_sep", "time_series_sep"),
    )
    dataframe_header_length: int = Field(
        2,
        description="Length of header of the dataframe",
        alias=AliasChoices(
            "dataframe_header_length", "time_series_header_length"
        ),
    )
    fillna: Optional[Any] = Field(
        "", description="Value to fill NaN values in the parsed dataframe."
    )
    # OVERRIDE
    mapping: Union[str, List[ABoxBaseMapping]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )

    # OVERRIDE
    @property
    def mapping_model(self) -> ABoxBaseMapping:
        """ABox Mapping Model for CSV Parser"""
        return ABoxBaseMapping

    # OVERRIDE
    @property
    def json_ld(self) -> "Dict[str, Any]":
        """
        Returns a JSON-LD representation of the CSV data in ABox mode.

        This method generates a JSON-LD object that describes the CSV data,
        including its metadata, dataframe data, and relationships between them.

        The returned JSON-LD object is in the format of a csvw:TableGroup,
        which contains one or more csvw:Table objects. Each csvw:Table object
        represents a table in the CSV data, and contains information about its
        columns, rows, and relationships to other tables.

        The JSON-LD object also includes context information, such as namespace
        prefixes and base URLs, to help with serialization and deserialization.

        Returns:
        Dict[str, Any]: A JSON-LD object representing the CSV data in ABox mode.
        """

        if not self.config.suppress_file_description:
            tables = []

            if self.general_metadata:
                meta_table = {
                    "@type": "csvw:Table",
                    "rdfs:label": "Metadata",
                    "csvw:row": [],
                }

                for n, mapping in enumerate(self.general_metadata):
                    if isinstance(mapping, QuantityGraph):
                        row = {
                            "@type": "csvw:Row",
                            "csvw:titles": {
                                "@type": "xsd:string",
                                "@value": mapping.key,
                            },
                            "csvw:rownum": {
                                "@type": "xsd:integer",
                                "@value": n,
                            },
                            "qudt:quantity": mapping.json_ld,
                        }
                        meta_table["csvw:row"].append(row)
                    elif isinstance(mapping, PropertyGraph):
                        row = {
                            "@type": "csvw:Row",
                            "csvw:titles": {
                                "@type": "xsd:string",
                                "@value": mapping.key,
                            },
                            "csvw:rownum": {
                                "@type": "xsd:integer",
                                "@value": n,
                            },
                            "csvw:describes": mapping.json_ld,
                        }
                        meta_table["csvw:row"].append(row)
                    else:
                        raise TypeError(
                            f"Mapping must be of type {QuantityGraph} or {PropertyGraph}, not {type(mapping)}"
                        )
                tables += [meta_table]

            if self.dataframe_metadata:
                column_schema = {"@type": "csvw:Schema", "csvw:column": []}
                tables += [
                    {
                        "@type": "csvw:Table",
                        "rdfs:label": "Dataframe",
                        "csvw:tableSchema": column_schema,
                    }
                ]
                for idx, mapping in enumerate(self.dataframe_metadata):
                    if isinstance(mapping, QuantityGraph):
                        entity = {"qudt:quantity": mapping.json_ld}
                    elif isinstance(mapping, PropertyGraph):
                        entity = {"dcterms:subject": mapping.json_ld}
                    else:
                        raise TypeError(
                            f"Mapping must be of type {QuantityGraph} or {PropertyGraph}, not {type(mapping)}"
                        )

                    if self.config.data_download_uri:
                        download_url = {
                            "dcterms:identifier": {
                                "@type": "xsd:anyURI",
                                "@value": urljoin(
                                    str(self.config.data_download_uri),
                                    f"column-{idx}",
                                ),
                            }
                        }
                    else:
                        download_url = {}

                    column = {
                        "@type": "csvw:Column",
                        "csvw:titles": {
                            "@type": "xsd:string",
                            "@value": mapping.key,
                        },
                        **entity,
                        "foaf:page": {
                            "@type": "foaf:Document",
                            "dcterms:format": {
                                "@type": "xsd:anyURI",
                                "@value": "https://www.iana.org/assignments/media-types/application/json",
                            },
                            "dcterms:type": {
                                "@type": "xsd:anyURI",
                                "@value": "http://purl.org/dc/terms/Dataset",
                            },
                            **download_url,
                        },
                    }
                    column_schema["csvw:column"].append(column)

            # flatten list if only one value exists
            if len(tables) == 1:
                tables = tables.pop()
            # make relation to csvw:table property
            if tables:
                csvw_tables = {"csvw:table": tables}
            else:
                csvw_tables = {}

            json_ld = {
                "@context": {
                    f"{self.config.prefix_name}": make_prefix(self.config),
                    "csvw": "http://www.w3.org/ns/csvw#",
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "dcat": "http://www.w3.org/ns/dcat#",
                    "xsd": "http://www.w3.org/2001/XMLSchema#",
                    "dcterms": "http://purl.org/dc/terms/",
                    "qudt": "http://qudt.org/schema/qudt/",
                    "csvw": "http://www.w3.org/ns/csvw#",
                    "foaf": "http://xmlns.com/foaf/spec/",
                },
                "@id": f"{self.config.prefix_name}:tableGroup",
                "@type": "csvw:TableGroup",
                **csvw_tables,
            }
        else:
            json_ld = {
                "@graph": [model.json_ld for model in self.general_metadata]
                + [model.json_ld for model in self.dataframe_metadata]
            }
        return json_ld

    # OVERRIDE
    @classmethod
    def _run_parser(
        cls,
        self: "CSVParser",
        datafile: StringIO,
        mapping: "List[ABoxBaseMapping]",
    ) -> None:
        """
        This function is responsible for parsing metadata, dataframe metadata, and dataframe data from a CSV file.

        It takes in three parameters:
        - `self`: The CSVParser instance.
        - `datafile`: The StringIO object containing the CSV data.
        - `mapping`: A list of ABoxBaseMapping instances that map the CSV data to the desired output format.

        The function returns None, but it populates the following instance variables:
        - `self._general_metadata`: A list of PropertyGraph or QuantityGraph instances representing the general metadata.
        - `self._dataframe_metadata`: A list of QuantityGraph instances representing the dataframe metadata.
        - `self._dataframe`: A pandas DataFrame containing the dataframe data.

        The function also raises ValueError if the `metadata_length` is greater than 0 but `metadata_sep` is not set.
        It raises TypeError if the unit for a key is not a string.
        It raises MappingMissmatchWarning if no match is found in the mapping for a key.
        """

        for model in mapping:
            if model.custom_relations:
                raise RuntimeError(
                    "Custom relations for CSV parser is currently not supported"
                )

        mapping = {model.key: model for model in mapping}

        dataframe: Union[pd.DataFrame, List[None]] = cls._parse_dataframe(
            self, datafile
        )
        if self.dropna:
            dataframe.dropna(inplace=True)
        datafile.seek(0)

        # iterate over general metadata
        self._general_metadata = []
        if self.metadata_length > 0:
            if not self.metadata_sep:
                raise ValueError(
                    "`metadata_length` is > 0 but `metadata_sep` is not set"
                )
            metadata = pd.read_csv(
                datafile,
                sep=self.metadata_sep,
                nrows=self.metadata_length,
                names=["key", "value", "unit"],
                header=None,
            )
            # remove unneeded characters
            metadata = metadata.map(
                lambda value: _replace(value, self.config.remove_from_datafile)
            )
            metadata.replace({np.nan: self.fillna}, inplace=True)
            for i, metadatum in metadata.iterrows():
                # get the match from the mapping
                mapping_match = mapping.get(metadatum.key)

                # only map the data if a match is found
                if mapping_match:
                    # get unit
                    unit = mapping_match.unit or metadatum.unit or None
                    if unit:
                        if not isinstance(unit, str):
                            raise TypeError(
                                f"""Unit `{unit}` for key `{metadatum.key}` is not a string.
                                Is it a bad mapping?"""
                            )
                        unit = _strip_unit(unit, self.config.remove_from_unit)

                    # instanciate model
                    model_data = {
                        "key": metadatum.key,
                        "value": metadatum.value,
                        "unit": unit,
                        "iri": mapping_match.iri,
                        "suffix": mapping_match.suffix,
                        "annotation": mapping_match.annotation or None,
                        "config": self.config,
                    }
                    if mapping_match.value_relation:
                        model_data[
                            "value_relation"
                        ] = mapping_match.value_relation
                    if model_data.get("unit"):
                        if mapping_match.unit_relation:
                            model_data[
                                "unit_relation"
                            ] = mapping_match.unit_relation
                        model = QuantityGraph(**model_data)
                    else:
                        model = PropertyGraph(**model_data)
                    self._general_metadata.append(model)
                else:
                    warnings.warn(
                        f"No match found in mapping for key `{metadatum.key}`",
                        MappingMissmatchWarning,
                    )

        # parse dataframe data and meta data
        self._dataframe_metadata = []
        self._dataframe = {}

        for key in dataframe:
            # get matching mapping
            mapping_match = mapping.get(key)

            if mapping_match:
                # get unit
                unit = (
                    mapping_match.unit
                    or (
                        dataframe[key].iloc[0]
                        if self.dataframe_header_length == 2
                        else None
                    )
                    or None
                )

                if unit:
                    if not isinstance(unit, str):
                        raise TypeError(
                            f"""Unit `{unit}` for key `{key}` is not a string.
                            Is it a bad mapping?"""
                        )
                    unit = _strip_unit(unit, self.config.remove_from_unit)

                # assign model
                model = QuantityGraph(
                    key=key,
                    unit=unit,
                    iri=mapping_match.iri,
                    suffix=mapping_match.suffix,
                    annotation=mapping_match.annotation or None,
                    config=self.config,
                )
                if mapping_match.unit_relation:
                    model.unit_relation = mapping_match.unit_relation

                # append model
                self.dataframe_metadata.append(model)

                # assign dataframe data
                self._dataframe[model.suffix] = dataframe[key][
                    self.dataframe_header_length - 1 :
                ].to_list()

            else:
                warnings.warn(
                    f"No match found in mapping for key `{key}`",
                    MappingMissmatchWarning,
                )
        # set dataframe as pd dataframe
        self._dataframe = pd.DataFrame.from_dict(
            self._dataframe, orient="index"
        ).transpose()
        # check if drop na:
        if self.dropna:
            self._dataframe.dropna(how="all", inplace=True)

    # OVERRIDE
    @classmethod
    def _load_data_file(cls, self: "CSVABoxParser") -> StringIO:
        """Load csv file"""
        return _load_data_file(self)

    @classmethod
    def _parse_dataframe(
        cls, self: "CSVParser", datafile: "StringIO"
    ) -> Union[pd.DataFrame, List[None]]:
        if self.dataframe_sep:
            response = pd.read_csv(
                datafile,
                encoding=self.config.encoding,
                sep=self.dataframe_sep,
                skiprows=self.metadata_length,
            )
            response = response.map(
                lambda value: _replace(value, self.config.remove_from_datafile)
            )
            response.columns = [
                _replace(column, self.config.remove_from_datafile)
                for column in response.columns
            ]
        else:
            warnings.warn(
                "`dataframe_sep` is not set. Any potential dataframe in the data file will be skipped.",
                ParserWarning,
            )
            response = []
        return response


class CSVParser(BaseFileParser):
    """Parser for CSV/TSV files"""

    # OVERRIDE
    @property
    def _abox_parser(self) -> CSVABoxParser:
        """Pydantic Model for CSV ABox parser"""
        return CSVABoxParser

    # OVERRIDE
    @property
    def _tbox_parser(self) -> CSVTBoxParser:
        """Pydantic Model for CSV TBox parser"""
        return CSVTBoxParser

    # OVERRIDE
    @property
    def media_type(self) -> str:
        """IANA Media type definition of the resource to be parsed."""
        return "http://www.iana.org/assignments/media-types/text/csv"



"""Data2rdf excel parser"""

import json
import os
import warnings
from typing import Any, Dict, List, Optional, Union
from urllib.parse import quote, urljoin

import pandas as pd
from jsonpath_ng import parse
from pydantic import Field

from data2rdf.models.graph import PropertyGraph, QuantityGraph
from data2rdf.models.mapping import (
    CustomRelationPropertySubgraph,
    CustomRelationQuantitySubgraph,
)
from data2rdf.utils import make_prefix
from data2rdf.warnings import MappingMissmatchWarning

from .utils import _value_exists

from data2rdf.parsers.base import (  # isort:skip
    ABoxBaseParser,
    BaseFileParser,
    TBoxBaseParser,
)
from data2rdf.parsers.utils import (  # isort:skip
    _make_tbox_classes,
    _make_tbox_json_ld,
    _strip_unit,
    _check_jsonpath,
)

from data2rdf.models.mapping import (  # isort:skip
    ABoxBaseMapping,
    CustomRelation,
    TBoxBaseMapping,
)


def _load_data_file(
    self: "Union[JsonABoxParser, JsonTBoxParser]",
) -> "List[Dict[str, Any]]":
    """Load json file"""
    if isinstance(self.raw_data, str):
        if os.path.isfile(self.raw_data):
            with open(self.raw_data, encoding=self.config.encoding) as file:
                content = json.load(file)
        else:
            content = json.loads(self.raw_data)

    if isinstance(self.raw_data, (list, dict)):
        content = self.raw_data
    if not isinstance(self.raw_data, (str, dict, list)):
        raise TypeError(
            "Raw data must be of type `str` for a file path or a `dict` for a parsed json."
        )
    return content


class JsonTBoxParser(TBoxBaseParser):
    """Parser for JSON in TBox mode"""

    # OVERRIDE
    mapping: Union[str, List[TBoxBaseMapping]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )

    # OVERRIDE
    @property
    def json_ld(self) -> "Dict[str, Any]":
        """Return JSON-LD in TBox mode"""
        return _make_tbox_json_ld(self)

    # OVERRIDE
    @property
    def mapping_model(self) -> TBoxBaseMapping:
        "TBox mapping model"
        return TBoxBaseMapping

    # OVERRIDE
    @classmethod
    def _run_parser(
        cls,
        self: "JsonTBoxParser",
        datafile: "List[Dict[str, Any]]",
        mapping: "Dict[str, TBoxBaseMapping]",
    ) -> None:
        """
        Runs the parser in TBox mode.

        Args:
            self: An instance of JsonTBoxParser.
            datafile: A list of dictionaries containing the data to be parsed.
            mapping: A dictionary containing the mapping of the data.

        Returns:
            None
        """

        df = pd.DataFrame(datafile)
        _make_tbox_classes(self, df, mapping)

    # OVERRIDE
    @classmethod
    def _load_data_file(cls, self: "JsonTBoxParser") -> "List[Dict[str, Any]]":
        return _load_data_file(self)


class JsonABoxParser(ABoxBaseParser):
    """Parser for JSON in ABox mode"""

    # OVERRIDE
    mapping: Union[str, List[ABoxBaseMapping]] = Field(
        ...,
        description="""File path to the mapping file to be parsed or
        a list with the mapping.""",
    )

    expand_array: bool = Field(
        False,
        description="""When enabled, the jsonpath pointing to arrays in the data
        will be iterated so that the mapping will be applied to each element of the array.""",
    )

    # OVERRIDE
    @property
    def mapping_model(self) -> ABoxBaseMapping:
        "ABox mapping model"
        return ABoxBaseMapping

    # OVERRIDE
    @property
    def json_ld(self) -> Dict[str, Any]:
        """
        Returns the JSON-LD representation of the parser's data.

        This method generates the JSON-LD representation of the parser's data,
        including the context, id, type, and members. The members are generated
        based on the general metadata and dataframe metadata.

        The method returns a dictionary containing the JSON-LD representation.

        :return: A dictionary containing the JSON-LD representation.
        :rtype: Dict[str, Any]
        """
        if not self.config.suppress_file_description:
            members = []

            triples = {
                "@context": {
                    f"{self.config.prefix_name}": make_prefix(self.config),
                    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                    "xsd": "http://www.w3.org/2001/XMLSchema#",
                    "dcterms": "http://purl.org/dc/terms/",
                    "qudt": "http://qudt.org/schema/qudt/",
                    "foaf": "http://xmlns.com/foaf/spec/",
                    "prov": "<http://www.w3.org/ns/prov#>",
                },
                "@id": f"{self.config.prefix_name}:Dictionary",
                "@type": "prov:Dictionary",
                "prov:hadDictionaryMember": members,
            }

            for mapping in self.general_metadata:
                if isinstance(mapping, QuantityGraph):
                    entity = {
                        "@type": "prov:KeyEntityPair",
                        "prov:pairKey": {
                            "@type": "xsd:string",
                            "@value": mapping.key,
                        },
                        "prov:pairEntity": {
                            "@type": "prov:Entity",
                            "qudt:quantity": mapping.json_ld,
                        },
                    }
                    members.append(entity)
                elif isinstance(mapping, PropertyGraph):
                    entity = {
                        "@type": "prov:KeyEntityPair",
                        "prov:pairKey": {
                            "@type": "xsd:string",
                            "@value": mapping.key,
                        },
                        "prov:pairEntity": {
                            "@type": "prov:Entity",
                            "dcterms:hasPart": mapping.json_ld,
                        },
                    }
                    members.append(entity)
                else:
                    raise TypeError(
                        f"Mapping must be of type {QuantityGraph} or {PropertyGraph}, not {type(mapping)}"
                    )

            for idx, mapping in enumerate(self.dataframe_metadata):
                if not isinstance(mapping, QuantityGraph):
                    raise TypeError(
                        f"Mapping must be of type {QuantityGraph}, not {type(mapping)}"
                    )

                if self.config.data_download_uri:
                    download_url = {
                        "dcterms:identifier": {
                            "@type": "xsd:anyURI",
                            "@value": urljoin(
                                str(self.config.data_download_uri),
                                f"column-{idx}",
                            ),
                        }
                    }
                else:
                    download_url = {}

                entity = {
                    "@type": "prov:KeyEntityPair",
                    "prov:pairKey": {
                        "@type": "xsd:string",
                        "@value": mapping.key,
                    },
                    "prov:pairEntity": {
                        "@type": "prov:Entity",
                        "qudt:quantity": mapping.json_ld,
                        "foaf:page": {
                            "@type": "foaf:Document",
                            "dcterms:format": {
                                "@type": "xsd:anyURI",
                                "@value": "https://www.iana.org/assignments/media-types/application/json",
                            },
                            "dcterms:type": {
                                "@type": "xsd:anyURI",
                                "@value": "http://purl.org/dc/terms/Dataset",
                            },
                            **download_url,
                        },
                    },
                }
                members.append(entity)
        else:
            triples = {
                "@graph": [model.json_ld for model in self.general_metadata]
                + [model.json_ld for model in self.dataframe_metadata]
            }

        return triples

    # OVERRIDE
    @classmethod
    def _load_data_file(cls, self: "JsonABoxParser") -> "Dict[str, Any]":
        """
        Class method for loading data file.

        Args:
            cls: The class of the parser.
            self: An instance of JsonABoxParser.

        Returns:
            Dict[str, Any]: The loaded data file.
        """
        return _load_data_file(self)

    # OVERRIDE
    @classmethod
    def _run_parser(
        cls,
        self: "JsonABoxParser",
        datafile: "Dict[str, Any]",
        mapping: "List[ABoxBaseMapping]",
    ) -> None:
        """
        Class method for parsing metadata, dataframe metadata,
        and dataframe from a given data file and mapping.

        Args:
            self: An instance of JsonABoxParser.
            datafile: A dictionary containing the data to be parsed.
            mapping: A list of ABoxJsonMapping objects defining the
                     mapping from the data to the ABox.

        Returns:
            None
        """
        self._general_metadata = []
        self._dataframe_metadata = []
        self._dataframe = {}
        for datum in mapping:
            subdataset = self._get_optional_subdataset(datafile, datum)

            if not datum.custom_relations:
                suffix = self._make_suffix_from_location(datum, subdataset)
                path = _check_jsonpath(datum.value_location)
                value_expression = parse(path)

                results = [
                    match.value for match in value_expression.find(subdataset)
                ]

                if len(results) == 0:
                    value = None
                    message = f"""Concept with key `{datum.key or path}`
                                    does not have a value at location `{path}`.
                                    Concept will be omitted in graph.
                                    """
                    warnings.warn(message, MappingMissmatchWarning)
                elif len(results) == 1:
                    value = results.pop()
                else:
                    value = results

                if isinstance(value, list) or _value_exists(value):
                    if datum.unit_location:
                        path_unit_location = _check_jsonpath(
                            datum.unit_location
                        )
                        unit_expression = parse(path_unit_location)

                        results = [
                            match.value
                            for match in unit_expression.find(subdataset)
                        ]

                        if len(results) == 0:
                            unit = None
                            message = f"""Concept with key `{datum.key or path_unit_location}`
                                            does not have a unit at location `{path_unit_location}`.
                                            Concept will be omitted in graph."""
                            warnings.warn(message, MappingMissmatchWarning)
                        elif len(results) == 1:
                            unit = results.pop()
                        else:
                            unit = None
                            message = f"""Concept with key `{datum.key or path_unit_location}`
                                        has multiple units at location `{path_unit_location}`.
                                        Concept will be omitted in graph."""
                            warnings.warn(message, MappingMissmatchWarning)

                    else:
                        unit = None

                    # decide which unit to take
                    unit = datum.unit or unit
                    if unit:
                        if not isinstance(unit, str):
                            raise TypeError(
                                f"""Unit `{unit}` for key `{datum.key}` is not a string.
                                Is it a bad mapping?"""
                            )
                        unit = _strip_unit(unit, self.config.remove_from_unit)

                    # make model
                    model_data = {
                        "key": datum.key or datum.value_location,
                        "iri": datum.iri,
                        "suffix": suffix,
                        "annotation": datum.annotation,
                        "config": self.config,
                    }
                    if datum.value_relation:
                        model_data["value_relation"] = datum.value_relation

                    # if we have a series and a unit and we are *not* expanding:
                    # * make a QuantityGraph with the unit
                    # * add the graph to the dataframe metadata
                    # * add the values of the series to the dataframe array
                    if (
                        isinstance(value, list)
                        and unit
                        and not self.expand_array
                    ):
                        model_data["unit"] = unit
                        if datum.unit_relation:
                            model_data["unit_relation"] = datum.unit_relation
                        model = QuantityGraph(**model_data)

                        self._dataframe[suffix] = value
                        self._dataframe_metadata.append(model)
                    # if we have a series in the form of a list and a unit and we are expanding:
                    # * iterate over the series
                    # * make a QuantityGraph with the unit and each iterated value
                    # * add the graph to the general metadata
                    elif (
                        isinstance(value, list) and unit and self.expand_array
                    ):
                        model_data["unit"] = unit
                        if datum.unit_relation:
                            model_data["unit_relation"] = datum.unit_relation
                        for val in value:
                            model = QuantityGraph(**model_data, value=val)
                            self._general_metadata.append(model)
                    # if we have a series and *no* unit and we are *not* expanding:
                    # * make a PropertyGraph
                    # * add the graph to the dataframe metadata
                    # * add the values of the series to the dataframe array
                    elif (
                        isinstance(value, list)
                        and not unit
                        and not self.expand_array
                    ):
                        model = PropertyGraph(
                            value_relation_type=datum.value_relation_type,
                            **model_data,
                        )
                        self._dataframe[suffix] = value
                        self._dataframe_metadata.append(model)
                    # if we have a series in the form of a list and *no* unit and we are expanding:
                    # * iterate over the series
                    # * make a PropertyGraph with each iterated value
                    # * add the graph to the general metadata
                    elif (
                        isinstance(value, list)
                        and not unit
                        and self.expand_array
                    ):
                        for val in value:
                            model = PropertyGraph(
                                value=val,
                                value_relation_type=datum.value_relation_type,
                                value_datatype=datum.value_datatype,
                                **model_data,
                            )
                            self._general_metadata.append(model)
                    # if we do *not* have a series but have a unit:
                    # * make a QuantityGraph with the unit and the value
                    # * add the graph to the general metadata
                    elif _value_exists(value) and unit:
                        model_data["value"] = value
                        model_data["unit"] = unit
                        if datum.unit_relation:
                            model_data["unit_relation"] = datum.unit_relation
                        model = QuantityGraph(**model_data)

                        self._general_metadata.append(model)
                    # if we do *not* have a series and *no* unit:
                    # * make a PropertyGraph with the value
                    # * add the graph to the general metadata
                    elif _value_exists(value) and not unit:
                        model = PropertyGraph(
                            value_relation_type=datum.value_relation_type,
                            value_datatype=datum.value_datatype,
                            value=value,
                            **model_data,
                        )
                        self._general_metadata.append(model)
                    else:
                        raise RuntimeError(
                            f"""Combination of data types not supported!
                            value: {value} ({type(value)})
                            unit: {unit}
                            expand array: {self.expand_array}"""
                        )

            else:
                for relation in datum.custom_relations:
                    if datum.source:
                        for sub in subdataset:
                            suffix = self._make_suffix_from_location(
                                datum, sub
                            )
                            self._make_custom_relation(
                                relation, sub, datum, suffix
                            )
                    else:
                        suffix = self._make_suffix_from_location(
                            datum, subdataset
                        )
                        self._make_custom_relation(
                            relation, subdataset, datum, suffix
                        )

        # set dataframe as pd dataframe
        self._dataframe = pd.DataFrame.from_dict(
            self._dataframe, orient="index"
        ).transpose()
        # check if drop na:
        if self.dropna:
            self._dataframe.dropna(how="all", inplace=True)

    def _get_optional_subdataset(
        self, datafile: Any, datum: ABoxBaseMapping
    ) -> Any:
        subdataset = None
        if datum.custom_relations and datum.source:
            path_source = _check_jsonpath(datum.source)
            value_expression = parse(path_source)
            results = [
                match.value for match in value_expression.find(datafile)
            ]
            if len(results) == 0:
                message = f"""Could not properly resolve location `{path_source}` for curstom relations."""
                warnings.warn(message, MappingMissmatchWarning)
            else:
                subdataset = results
        else:
            subdataset = datafile
        return subdataset or datafile

    def _make_custom_relation(
        self,
        relation: CustomRelation,
        subdataset: Any,
        datum: ABoxBaseMapping,
        suffix: str,
    ) -> None:
        path_object_location = _check_jsonpath(relation.object_location)
        value_expression = parse(path_object_location)

        results = [match.value for match in value_expression.find(subdataset)]

        if len(results) == 0:
            value = None
            message = f"""Concept with for iri `{datum.iri}`
                            does not have a value at location `{path_object_location}`.
                            Concept will be omitted in graph.
                            """
            warnings.warn(message, MappingMissmatchWarning)
        elif len(results) == 1:
            value = results.pop()
        else:
            value = results

        if isinstance(
            relation.object_data_type,
            (CustomRelationPropertySubgraph, CustomRelationQuantitySubgraph),
        ):
            if isinstance(value, list):
                for val in value:
                    self._make_subgraph(relation, datum, val, suffix)
            elif _value_exists(value):
                self._make_subgraph(relation, datum, value, suffix)
            else:
                message = f"""Concept with for iri `{datum.iri}`
                                does not have a value at location `{relation.object_location}`.
                                Concept will be omitted in graph.
                                """
                warnings.warn(message, MappingMissmatchWarning)

        else:
            if isinstance(value, list):
                for val in value:
                    self._make_property_graph(
                        val,
                        datum.iri,
                        suffix,
                        **relation.model_dump(exclude={"object_location"}),
                    )
            elif _value_exists(value):
                self._make_property_graph(
                    value,
                    datum.iri,
                    suffix,
                    **relation.model_dump(exclude={"object_location"}),
                )
            else:
                message = f"""Concept with for iri `{datum.iri}`
                                does not have a value at location `{relation.object_location}`.
                                Concept will be omitted in graph.
                                """
                warnings.warn(message, MappingMissmatchWarning)

    def _make_subgraph(
        self,
        relation: CustomRelation,
        datum: ABoxBaseMapping,
        value: Any,
        suffix: str,
    ) -> None:
        if relation.object_data_type.concatenate:
            iri = str(relation.object_data_type.iri)
            iri = iri if iri.endswith("/") else iri + "/"
            value = urljoin(iri, str(value))
            self._make_property_graph(
                value,
                datum.iri,
                suffix,
                relation=relation.relation,
                relation_type="object_property",
            )
        else:
            if isinstance(
                relation.object_data_type, CustomRelationPropertySubgraph
            ):
                Model = PropertyGraph
            else:
                Model = QuantityGraph
            model = Model(
                value=value, **relation.object_data_type.model_dump()
            )
            model.suffix += "_" + suffix
            self._make_property_graph(
                model,
                datum.iri,
                suffix,
                relation=relation.relation,
                relation_type="object_property",
            )

    def _make_property_graph(
        self,
        value: Any,
        iri: str,
        suffix: str,
        relation: Optional[str] = None,
        relation_type: Optional[str] = None,
        object_data_type: Optional[str] = None,
    ) -> None:
        model = PropertyGraph(
            value_relation=relation,
            value_relation_type=relation_type,
            value_datatype=object_data_type,
            value=value,
            iri=iri,
            suffix=suffix,
            config=self.config,
        )
        self._general_metadata.append(model)

    def _make_suffix_from_location(
        self, datum: ABoxBaseMapping, subdataset: Any
    ) -> str:
        if datum.suffix_from_location:
            path_suffix = _check_jsonpath(datum.suffix)
            value_expression = parse(path_suffix)
            results = [
                match.value for match in value_expression.find(subdataset)
            ]

            if len(results) == 0 or len(results) > 1:
                suffix = path_suffix
                message = f"""Could not properly resolve suffix location `{path_suffix}`
                                Will use the location itself as suffix.
                            """
                warnings.warn(message, MappingMissmatchWarning)
            else:
                suffix = results.pop()
        else:
            suffix = datum.suffix

        suffix = quote(suffix)

        return suffix


class JsonParser(BaseFileParser):
    """
    Parses a data file of type json
    """

    # OVERRIDE
    @property
    def media_type(self) -> str:
        """IANA Media type definition of the resource to be parsed."""
        return "https://www.iana.org/assignments/media-types/application/json"

    # OVERRIDE
    @property
    def _abox_parser(self) -> JsonABoxParser:
        """Pydantic Model for Joson ABox parser"""
        return JsonABoxParser

    # OVERRIDE
    @property
    def _tbox_parser(self) -> JsonTBoxParser:
        """Pydantic Model for Excel TBox parser"""
        return JsonTBoxParser



"""Data2RDF model utilities"""

import ast
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict


def is_integer(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def is_bool(s):
    try:
        ast.literal_eval(s)
        return True
    except Exception:
        return False


def is_uri(s):
    try:
        if str(s).startswith("http://") or str(s).startswith("https://"):
            return True
    except Exception:
        return False


def detect_datatype(value) -> "Dict[str, Any]":
    """Return json with value definition"""
    if is_integer(value):
        dtype = "xsd:integer"
        value = int(value)
    elif is_float(value):
        dtype = "xsd:float"
        value = float(value)
    elif is_bool(value):
        dtype = "xsd:boolean"
        value = ast.literal_eval(value)
    elif is_uri(value):
        dtype = "xsd:anyURI"
        value = str(value)
    elif isinstance(value, str):
        dtype = "xsd:string"
    else:
        raise TypeError(
            f"Datatype of value `{value}` ({type(value)}) cannot be mapped to xsd."
        )

    return {"@type": dtype, "@value": value}


def apply_datatype(value: "Any", datatype: str) -> "Dict[str, Any]":
    """
    Converts the input value to the specified datatype and returns a dictionary
    with the datatype and the converted value.

    Args:
        value (Any): The value to be converted.
        datatype (str): The target datatype as a string. Supported datatypes
                        are "integer", "float", "bool"/"boolean", "anyURI", and "string".

    Returns:
        Dict[str, Any]: A dictionary containing the datatype under the "@type" key
                        and the converted value under the "@value" key.

    Raises:
        TypeError: If the provided datatype is not supported.
    """
    if datatype == "integer":
        value = int(value)
    elif datatype == "float":
        value = float(value)
    elif datatype == "bool":
        value = ast.literal_eval(value)
    elif datatype == "anyURI":
        value = str(value)
    elif datatype == "string":
        value = str(value)
    else:
        raise TypeError(
            f"""Datatype of value `{value}` ({datatype}) currently not supported.
            Supported datatypes: integer, float, boolean, anyURI, string"""
        )

    return {"@type": f"xsd:{datatype}", "@value": value}



"""Test the csv parser when rows might be empty"""

import os

from ..utils import dsms_schema

test_folder = os.path.dirname(os.path.abspath(__file__))
working_folder = os.path.join(test_folder, "input")
output_folder = os.path.join(test_folder, "output")

mapping_folder = os.path.join(working_folder, "mapping")
raw_data = os.path.join(working_folder, "data", "data.csv")
expected = os.path.join(output_folder, "output_csv_parser.ttl")

parser_args = {
    "dataframe_sep": ";",
    "metadata_length": 0,
    "dataframe_header_length": 1,
    "drop_na": False,
}

columns = [
    "Temperature",
    "ThermalExpansionCoefficient",
    "SpecificHeatCapacity",
    "ModulusOfElasticity",
    "PoissonRatio",
    "ThermalConductivity",
    "MassDensity",
]

config = {"graph_identifier": "https://www.example.org"}


def test_csv_nan_vals() -> None:
    from rdflib import Graph

    from data2rdf import QuantityGraph
    from data2rdf.parsers import CSVParser

    parser = CSVParser(
        raw_data=raw_data,
        mapping=os.path.join(mapping_folder, "mapping.json"),
        parser_args=parser_args,
        config=config,
    )

    expected_graph = Graph()
    expected_graph.parse(expected)

    assert len(parser.general_metadata) == 0

    assert len(parser.dataframe_metadata) == 7
    for row in parser.dataframe_metadata:
        assert isinstance(row, QuantityGraph)

    assert len(parser.dataframe.columns) == 7
    assert sorted(list(parser.dataframe.columns)) == sorted(columns)

    for name, column in parser.dataframe.items():
        assert len(column) == 31

    assert parser.graph.isomorphic(expected_graph)
    assert parser.to_dict(schema=dsms_schema) == {}
    assert parser.to_dict() == {}



"""Models for graph construction from semantic concepts"""

import warnings
from typing import Any, Dict, List, Optional, Union

from data2rdf.qudt.utils import _get_qudt_label_and_symbol, _get_query_match
from data2rdf.utils import make_prefix, split_namespace
from data2rdf.warnings import ParserWarning

from data2rdf.models.utils import (  # isort:skip
    apply_datatype,
    detect_datatype,
    is_float,
    is_integer,
)

from data2rdf.models.base import (  # isort:skip
    BasicGraphModel,
    BasicSuffixModel,
    RelationType,
    BaseConfigModel,
)

from pydantic import (  # isort:skip
    AnyUrl,
    AliasChoices,
    BaseModel,
    Field,
    ValidationInfo,
    field_validator,
    model_validator,
)


class ValueRelationMapping(BaseModel):
    """Mapping between a object/data/annotation property and a value resolved from a location in the data file"""

    value: Union[str, int, float, bool, AnyUrl] = Field(
        ...,
        description="""Value resolved from the data file.""",
    )
    relation: Union[str, AnyUrl] = Field(
        ...,
        description="""Object/Data/Annotation property for the value
        resolving from `key` of this model""",
    )
    datatype: Optional[str] = Field(
        None, description="XSD Datatype of the value"
    )


class ClassTypeGraph(BasicGraphModel):
    """Graph of a potential concept or class in the T Box."""

    suffix: str = Field(
        ...,
        description="""Value of the suffix of the
        ontological class to be used""",
    )
    rdfs_type: str = Field(
        "owl:Class", description="rdfs:type for this concept"
    )
    annotation_properties: Optional[List[ValueRelationMapping]] = Field(
        None, description="Mappings for Annotations Properties"
    )
    object_properties: Optional[List[ValueRelationMapping]] = Field(
        None, description="Mappings for Object Properties"
    )
    data_properties: Optional[List[ValueRelationMapping]] = Field(
        None, description="Mappings for Data Properties"
    )

    rdfs_properties: Optional[List[ValueRelationMapping]] = Field(
        None, description="Mappings for rdfs:Properties"
    )

    # OVERRIDE
    @property
    def json_ld(self) -> "Dict[str, Any]":
        annotations = {
            model.relation: (
                apply_datatype(model)
                if model.datatype
                else detect_datatype(str(model.value))
            )
            for model in self.annotation_properties
        }
        datatypes = {
            model.relation: (
                apply_datatype(model.value, model.datatype)
                if model.datatype
                else detect_datatype(str(model.value))
            )
            for model in self.data_properties
        }
        properties = {
            model.relation: (
                apply_datatype(model)
                if model.datatype
                else detect_datatype(str(model.value))
            )
            for model in self.rdfs_properties
        }
        objects = {
            model.relation: {"@id": model.value}
            for model in self.object_properties
        }
        return {
            "@context": {
                "owl": "http://www.w3.org/2002/07/owl#",
                "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
                "dcterms": "http://purl.org/dc/terms/",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                f"{self.config.prefix_name}": make_prefix(self.config),
            },
            "@id": f"{self.config.prefix_name}:{self.suffix}",
            "@type": str(self.rdfs_type),
            **annotations,
            **datatypes,
            **objects,
            **properties,
        }


class MeasurementUnit(BaseConfigModel):
    iri: Union[str, AnyUrl] = Field(
        ...,
        description="Ontological IRI related to the measurement unit",
    )
    label: Optional[str] = Field(
        None,
        description="Label of the measurement unit",
    )
    symbol: Optional[str] = Field(
        None,
        description="Symbol of the measurement unit",
    )
    namespace: Optional[str] = Field(
        None,
        description="Namespace of the measurement unit",
    )

    @model_validator(mode="after")
    @classmethod
    def validate_measurement_unit(cls, self) -> "MeasurementUnit":
        unit = _get_qudt_label_and_symbol(
            self.iri, self.config.qudt_units, self.config.language
        )
        if not self.label and "label" in unit:
            self.label = unit["label"]
        if not self.symbol and "symbol" in unit:
            self.symbol = unit["symbol"]
        if not self.namespace:
            self.namespace = split_namespace(self.iri)
        return self


class QuantityGraph(BasicGraphModel, BasicSuffixModel):
    """Quantity with or without a discrete value and a unit
    E.g. a quantity with a single value and unit _or_
    a quantity describing a column of a dataframe or table with a unit."""

    unit: Optional[Union[str, AnyUrl]] = Field(
        None, description="QUDT Symbol or any other IRI for the unit mapping"
    )
    value: Optional[Union[int, float, str]] = Field(
        None, description="Value of the quantity"
    )

    unit_relation: Optional[Union[str, AnyUrl]] = Field(
        "qudt:hasUnit",
        description="""Object property for mapping the IRI
         of the unit to the individual.""",
    )

    value_relation: Optional[Union[str, AnyUrl]] = Field(
        "qudt:value",
        description="""Data property
        for mapping the data value to the individual.""",
    )

    measurement_unit: Optional[MeasurementUnit] = Field(
        None,
        description="Detailed QUDT Measurement Unit specification",
        alias=AliasChoices(
            "measurement_unit", "measurementunit", "measurementUnit"
        ),
    )

    @field_validator("value", mode="after")
    @classmethod
    def validate_value(
        cls, value: Union[int, float, str]
    ) -> Union[int, float]:
        if isinstance(value, str) and is_integer(value):
            value = int(value)
        elif isinstance(value, str) and is_float(value):
            value = float(value)
        elif isinstance(value, str):
            warnings.warn(
                f"Cannot type case value from str into float or int: {value}",
                ParserWarning,
            )
        return value

    @field_validator("unit", mode="after")
    @classmethod
    def validate_unit(
        cls, value: Union[str, AnyUrl], info: ValidationInfo
    ) -> Optional[AnyUrl]:
        config = info.data.get("config")
        if isinstance(value, str):
            if not (value.startswith("https:") or value.startswith("http:")):
                match = _get_query_match(value, config.qudt_units)
                if len(match) == 0:
                    warnings.warn(
                        f"No QUDT Mapping found for unit with symbol `{value}`."
                    )
                    value = None
                elif len(match) > 1:
                    warnings.warn(
                        f"Multiple QUDT Mappings found for unit with symbol `{value}`."
                    )
                    value = match.pop()
                else:
                    value = match.pop()
        elif isinstance(value, AnyUrl):
            value = str(value)
        return value

    @model_validator(mode="after")
    @classmethod
    def validate_quantity_graph(cls, self) -> "QuantityGraph":
        if not self.measurement_unit and self.unit:
            self.measurement_unit = MeasurementUnit(iri=self.unit)
        if self.measurement_unit and not self.unit:
            self.unit = self.measurement_unit.iri
        return self

    @property
    def json_ld(self) -> Dict[str, Any]:
        """Return dict of json-ld for graph"""
        return {
            "@context": {
                f"{self.config.prefix_name}": make_prefix(self.config),
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
                "qudt": "http://qudt.org/schema/qudt/",
            },
            "@id": f"{self.config.prefix_name}:{self.suffix}",
            "@type": (
                [str(iri) for iri in self.iri]
                if isinstance(self.iri, list)
                else str(self.iri)
            ),
            **self.unit_json,
            **self.value_json,
        }

    @property
    def unit_json(self) -> "Dict[str, Any]":
        """Return json with unit definition"""
        if self.unit:
            value = {
                self.unit_relation: {
                    "@value": self.unit,
                    "@type": "xsd:anyURI",
                }
            }
        else:
            value = {}
        return value

    @property
    def value_json(self) -> "Dict[str, Any]":
        """Return json with value definition"""
        if self.value:
            value = {self.value_relation: detect_datatype(str(self.value))}
        else:
            value = {}
        return value


class PropertyGraph(BasicGraphModel, BasicSuffixModel):
    """Mapping for an individual with arbitrary property. E.g. the
    name of a tester or a testing facility. The value must not have a
    discrete value but can also be a reference to a column in a table or
    dataframe."""

    value: Optional[
        Union[str, int, float, bool, AnyUrl, "PropertyGraph", "QuantityGraph"]
    ] = Field(None, description="Value of the property")
    annotation: Optional[Union[str, AnyUrl]] = Field(
        None, description="Base IRI with which the value shall be concatenated"
    )
    value_relation: Optional[Union[str, AnyUrl]] = Field(
        "rdfs:label",
        description="""Data or annotation property
        for mapping the data value to the individual.""",
        alias=AliasChoices("relation", "value_relation", "valuerelation"),
    )
    value_relation_type: Optional[RelationType] = Field(
        None,
        description="Type of the semantic relation used in the mappings",
        alias=AliasChoices(
            "value_relation_type",
            "value_relationtype",
            "relation_type",
            "relationtype",
        ),
    )
    value_datatype: Optional[str] = Field(
        None,
        description="In case of an annotation or data property, this field indicates the XSD Datatype of the value",
        alias=AliasChoices(
            "value_datatype", "value_data_type", "datatype", "data_type"
        ),
    )

    @field_validator("annotation")
    @classmethod
    def validate_annotation(cls, value: AnyUrl) -> AnyUrl:
        """Make sure that there are not blank spaces in the IRI"""
        if value:
            value = AnyUrl(str(value).strip())
        return value

    @model_validator(mode="after")
    @classmethod
    def validate_value(cls, self: "PropertyGraph") -> "PropertyGraph":
        """
        Validate value of a property graph.

        In case the value is a property graph or a quantity graph, make sure that
        the config is set correctly.
        """
        if isinstance(self.value, (PropertyGraph, QuantityGraph)):
            self.value.config = self.config
        return self

    @model_validator(mode="after")
    @classmethod
    def validate_property_graph(cls, self: "PropertyGraph") -> "PropertyGraph":
        """Validate property graph in order to generate annotations"""
        if self.annotation:
            if str(self.annotation).endswith(self.config.separator):
                self.annotation = str(self.annotation) + self.value
            else:
                self.annotation = (
                    str(self.annotation) + self.config.separator + self.value
                )
        return self

    @property
    def json_ld(self) -> Dict[str, Any]:
        """Return dict of json-ld for graph"""
        return {
            "@context": {
                f"{self.config.prefix_name}": make_prefix(self.config),
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "xsd": "http://www.w3.org/2001/XMLSchema#",
            },
            "@id": f"{self.config.prefix_name}:{self.suffix}",
            **self.value_json,
            **self.types_json,
        }

    @property
    def value_json(self) -> "Optional[Dict[str, str]]":
        if not isinstance(self.value, type(None)):
            if self.value_relation_type != RelationType.OBJECT_PROPERTY:
                if not self.value_datatype:
                    spec = detect_datatype(str(self.value))
                else:
                    spec = apply_datatype(self.value, self.value_datatype)
                response = {self.value_relation: spec}
            else:
                if isinstance(self.value, (PropertyGraph, QuantityGraph)):
                    response = {self.value_relation: self.value.json_ld}
                else:
                    response = {self.value_relation: {"@id": str(self.value)}}
        else:
            response = {}
        return response

    @property
    def types_json(self) -> "Dict[str, Any]":
        """Dict of json-ld for class types of the individual"""
        if self.annotation:
            types = {
                "@type": [
                    (
                        [str(iri) for iri in self.iri]
                        if isinstance(self.iri, list)
                        else str(self.iri)
                    ),
                    self.annotation,
                ]
            }
        else:
            types = {
                "@type": [
                    (
                        [str(iri) for iri in self.iri]
                        if isinstance(self.iri, list)
                        else str(self.iri)
                    )
                ]
            }
        return types



"""Data2RDF parsers."""

from enum import Enum

from .csv import CSVParser
from .excel import ExcelParser
from .json import JsonParser

Parser = Enum(
    "Parser",
    {"csv": CSVParser, "excel": ExcelParser, "json": JsonParser},
)

__all__ = ["Parser"]
